/*!\file main.h
 * \brief public funktions and variables
 * \author F.Satzger
 *
 * \date 22. Oktober 2010, 10:59
 */

#ifndef _MAIN_H
#define	_MAIN_H


/*! \brief Sets windows version information for some Apis.
*/
#define WINVER 0x0501

/*! \brief Sets windows version information for some Apis.
 */

#define _WIN32_IE 0x600
#include <initguid.h>
#include <stdlib.h>
#include <windows.h>
#include <winuser.h>
#include <stdio.h>
#include <setupapi.h>
#include <ddk/batclass.h>
#include "mainres.h"

/*! \brief Root namespace
 */
namespace BatteryMonitor{

/*! \brief The window procedure.
 *
 * Here, all window logic is done such as mouse handling and so on.
 * \see Windows API for further information
 */
LRESULT CALLBACK wndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

/*! \brief Adds an icon to the taskbar.
 *
 * \param [in] hwnd The window handle of the window whose window procedure
 * receives the events occuring.
 * \see Windows API
 */
void addNotifyIcon(HWND hwnd);

/*! \brief Error output.
 *
 * If BatteryMonitor is compiled in debug configuration (with a DEBUG defined) , a messagebox
 * containing the error text is shown.
 */
void errOut(void);

/*! \brief Shows a balloon tip with battery information.
 */
void showBatteryInfo(HWND hwnd);

/*! \brief Tries to shut down the computer.
 *
 * This function is called if remaining battery lifetime is less than 5 minutes.
 * \see addNotifyIcon(HWND hwnd)
 */
void shutdown(void);

/*! \brief Retrieves the path of the battery device.
 *
 * This function retrieves the path of the battery device and stores it
 * into #devicepath.
 */
void initDevice(void);

/*! \brief Shows the main window.
 *
 * This funktion creates a window of the class stored in #classname. This
 * window is automatically set click-through, transparent and topmost.
 * \param [in] nCmdShow Can override the default visibility-parameters
 * \return The window handle of the newly created window
 * \see Windows API
 */
HWND setupWindow(int nCmdShow);

/*! \brief registers the window class stored in #classname.
 *
 * \return The error code of RegisterClassEx, the corresponding windows api call.
 */
long setupWindowClass(void);

/*! \brief Memory and handler cleanups, taskbar icon removal.
 *
 * \param [in] hwnd The window handle of the window associated with the taskbar icon.
 */
void cleanUp(HWND hwnd);

/*! \brief retrieves battery informations from the system.
 *
 * This funktion retrieves battery information from the system.
 * If any error occurs, errOut() will be called and the DesignedCapacity member
 * of the returned struct is set to 0.
 * \return Battery informations
 */
BATTERY_INFORMATION getBatteryInformation() ;


/*! \brief Window message for updating battery status.
 */
#define UPDATE_BATTERY 1

/*! \brief Path to battery device.
 *
 * The batteryinformation can be read from a pseudo-file, which path is
 * stored here.
 * \see Windows API
 */
char* devicepath;

/*! \var HBRUSH red
 *  \brief Red brush
 */
/*! \var HBRUSH blue
 *  \brief Blue brush
 */
/*! \var HBRUSH yellow
 *  \brief Yellow brush
 */
/*! \var HBRUSH green
 *  \brief Green brush
 */
HBRUSH red,yellow,blue,green;

/*! \brief Application instance.
 * \see Windows API
 */
HINSTANCE instance;

 /*! \brief Battery charge.
  *
  * Battery charge, positive if battery is discharging,
  * negative if charging. (e.g. if 75% and changing, battery is 0.75).
  * If battery status cannot be determined, battery is 2.
  */
double battery=0;

/*! \brief Remaining battery lifetime in seconds.
 *
 * Remaining battery lifetime in seconds. -1 is status cannot be determined.
 */
long batterylifetime=-1;

/*! \brief Name of registered window class.
 *
 * \see Windows API
 */
char classname[14];

/*! \brief Window message for taskbarcreated-event.
 * 
 * This message must be registered with windows, so that the application gets
 * informed if the taskbar is (re-)created. This is necessary to update the
 * taskbar-icon.
 */
UINT WM_TASKBARCREATED=0;

/*! \brief Indicates, if shutdown was invoked by BatteryMonitor
 */
bool shutdown_enabled=false;

}
#endif	/* _MAIN_H */

